@@ -143,11 +143,6 @@ version (void)
 static void
 meta_print_compilation_info (void)
 {
-#ifdef HAVE_SHAPE
-  meta_verbose ("Compiled with shape extension\n");
-#else
-  meta_verbose ("Compiled without shape extension\n");
-#endif
 #ifdef HAVE_XINERAMA
   meta_topic (META_DEBUG_XINERAMA, "Compiled with Xinerama extension\n");
 #else
@@ -163,11 +158,6 @@ meta_print_compilation_info (void)
 #else
   meta_topic (META_DEBUG_XINERAMA, " (not using Solaris Xinerama)\n");
 #endif
-#ifdef HAVE_XSYNC
-  meta_verbose ("Compiled with sync extension\n");
-#else
-  meta_verbose ("Compiled without sync extension\n");
-#endif
 #ifdef HAVE_RANDR
   meta_verbose ("Compiled with randr extension\n");
 #else
@@ -649,7 +639,8 @@ prefs_changed_callback (MetaPreference pref,
   switch (pref)
     {
     case META_PREF_THEME:
-      meta_ui_set_current_theme (meta_prefs_get_theme (), FALSE);
+    case META_PREF_COMPOSITING_MANAGER:
+      meta_ui_set_current_theme (meta_prefs_get_theme (), TRUE);
       meta_display_retheme_all ();
       break;
 
