@@ -100,6 +100,8 @@ set_supported_hint (MetaScreen *screen)
 #include "atomnames.h"
 #undef item
 #undef EWMH_ATOMS_ONLY
+
+    screen->display->atom__GTK_FRAME_EXTENTS,
   };
 
   XChangeProperty (screen->display->xdisplay, screen->xroot,
@@ -1211,6 +1213,49 @@ meta_screen_update_cursor (MetaScreen *screen)
   XFreeCursor (screen->display->xdisplay, xcursor);
 }
 
+#define MAX_PREVIEW_SIZE 150.0
+
+static GdkPixbuf *
+get_window_pixbuf (MetaWindow *window,
+                   int        *width,
+                   int        *height)
+{
+  Pixmap pmap;
+  GdkPixbuf *pixbuf, *scaled;
+  double ratio;
+
+  pmap = meta_compositor_get_window_pixmap (window->display->compositor,
+                                            window);
+  if (pmap == None)
+    return NULL;
+
+  pixbuf = meta_ui_get_pixbuf_from_pixmap (pmap);
+  if (pixbuf == NULL) 
+    return NULL;
+
+  *width = gdk_pixbuf_get_width (pixbuf);
+  *height = gdk_pixbuf_get_height (pixbuf);
+
+  /* Scale pixbuf to max dimension MAX_PREVIEW_SIZE */
+  if (*width > *height)
+    {
+      ratio = ((double) *width) / MAX_PREVIEW_SIZE;
+      *width = (int) MAX_PREVIEW_SIZE;
+      *height = (int) (((double) *height) / ratio);
+    }
+  else
+    {
+      ratio = ((double) *height) / MAX_PREVIEW_SIZE;
+      *height = (int) MAX_PREVIEW_SIZE;
+      *width = (int) (((double) *width) / ratio);
+    }
+
+  scaled = gdk_pixbuf_scale_simple (pixbuf, *width, *height,
+                                    GDK_INTERP_BILINEAR);
+  g_object_unref (pixbuf);
+  return scaled;
+}
+                                         
 void
 meta_screen_ensure_tab_popup (MetaScreen      *screen,
                               MetaTabList      list_type,
@@ -1243,23 +1288,53 @@ meta_screen_ensure_tab_popup (MetaScreen      *screen,
     {
       MetaWindow *window;
       MetaRectangle r;
+      GdkPixbuf *win_pixbuf;
+      int width, height;
 
       window = tmp->data;
-
-      /* Metacity used to generated a thumbnail of the window contents
-       * using gdk_pixbuf_simple_scale which is so slow with bilinear
-       * interpolation that it has been removed.
-       *
-       * The cairo downscale could have been an alternative, but the
-       * result is terrible ATM, so it is simply better to have the
-       * same user story as the non-composited case (no thumbnail).
-       *
-       * https://bugzilla.gnome.org/show_bug.cgi?id=504729
-       */
-
+      
       entries[i].key = (MetaTabEntryKey) window->xwindow;
       entries[i].title = window->title;
-      entries[i].icon = g_object_ref (window->icon);
+
+      win_pixbuf = NULL;
+      if (meta_prefs_get_alt_tab_thumbnails ())
+        win_pixbuf = get_window_pixbuf (window, &width, &height);
+
+      if (win_pixbuf == NULL)
+        entries[i].icon = g_object_ref (window->icon);
+      else
+        {
+          GdkPixbuf *scaled;
+          int icon_width, icon_height, t_width, t_height;
+
+#define ICON_SIZE 32
+#define ICON_OFFSET 6
+
+          scaled = gdk_pixbuf_scale_simple (window->icon,
+                                            ICON_SIZE, ICON_SIZE,
+                                            GDK_INTERP_BILINEAR);
+
+          icon_width = gdk_pixbuf_get_width (scaled);
+          icon_height = gdk_pixbuf_get_height (scaled);
+
+          t_width = width + ICON_OFFSET;
+          t_height = height + ICON_OFFSET;
+
+          entries[i].icon = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8,
+                                            t_width, t_height);
+          gdk_pixbuf_fill (entries[i].icon, 0x00000000);
+          gdk_pixbuf_copy_area (win_pixbuf, 0, 0, width, height,
+                                entries[i].icon, 0, 0);
+          g_object_unref (win_pixbuf);
+          gdk_pixbuf_composite (scaled, entries[i].icon,
+                                t_width - icon_width, t_height - icon_height,
+                                icon_width, icon_height,
+                                t_width - icon_width, t_height - icon_height, 
+                                1.0, 1.0, GDK_INTERP_BILINEAR, 255);
+
+          g_object_unref (scaled);
+        }
+                                
       entries[i].blank = FALSE;
       entries[i].hidden = !meta_window_showing_on_its_workspace (window);
       entries[i].demands_attention = window->wm_state_demands_attention;
